@@ -283,7 +283,7 @@ static int handle_udp_read(struct osmo_fd *bfd)
 	if (other_conn) {
 		/* enqueue the message for TX on the respective FD */
 		msgb_enqueue(&other_conn->tx_queue, msg);
-		other_conn->fd.when |= BSC_FD_WRITE;
+		osmo_fd_write_enable(&other_conn->fd);
 	} else
 		msgb_free(msg);
 
@@ -293,7 +293,7 @@ static int handle_udp_read(struct osmo_fd *bfd)
 static int handle_udp_write(struct osmo_fd *bfd)
 {
 	/* not implemented yet */
-	bfd->when &= ~BSC_FD_WRITE;
+	osmo_fd_write_disable(bfd);
 
 	return -EIO;
 }
@@ -303,9 +303,9 @@ static int udp_fd_cb(struct osmo_fd *bfd, unsigned int what)
 {
 	int rc = 0;
 
-	if (what & BSC_FD_READ)
+	if (what & OSMO_FD_READ)
 		rc = handle_udp_read(bfd);
-	if (what & BSC_FD_WRITE)
+	if (what & OSMO_FD_WRITE)
 		rc = handle_udp_write(bfd);
 
 	return rc;
@@ -840,7 +840,7 @@ static int handle_tcp_read(struct osmo_fd *bfd)
 		/* enqueue packet towards BSC */
 		msgb_enqueue(&bsc_conn->tx_queue, msg);
 		/* mark respective filedescriptor as 'we want to write' */
-		bsc_conn->fd.when |= BSC_FD_WRITE;
+		osmo_fd_write_enable(&bsc_conn->fd);
 	} else {
 		logp_ipbc_uid(DLINP, LOGL_INFO, ipbc, bfd->priv_nr >> 8);
 		LOGPC(DLINP, LOGL_INFO, "Dropping packet from %s, "
@@ -869,7 +869,7 @@ static int handle_tcp_write(struct osmo_fd *bfd)
 
 	/* get the next msg for this timeslot */
 	if (llist_empty(&ipc->tx_queue)) {
-		bfd->when &= ~BSC_FD_WRITE;
+		osmo_fd_write_disable(bfd);
 		return 0;
 	}
 	lh = ipc->tx_queue.next;
@@ -897,12 +897,12 @@ static int proxy_ipaccess_fd_cb(struct osmo_fd *bfd, unsigned int what)
 {
 	int rc = 0;
 
-	if (what & BSC_FD_READ) {
+	if (what & OSMO_FD_READ) {
 		rc = handle_tcp_read(bfd);
 		if (rc < 0)
 			return rc;
 	}
-	if (what & BSC_FD_WRITE)
+	if (what & OSMO_FD_WRITE)
 		rc = handle_tcp_write(bfd);
 
 	return rc;
@@ -917,7 +917,7 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 	struct sockaddr_in sa;
 	socklen_t sa_len = sizeof(sa);
 
-	if (!(what & BSC_FD_READ))
+	if (!(what & OSMO_FD_READ))
 		return 0;
 
 	ret = accept(listen_bfd->fd, (struct sockaddr *) &sa, &sa_len);
@@ -936,11 +936,7 @@ static int listen_fd_cb(struct osmo_fd *listen_bfd, unsigned int what)
 	}
 
 	bfd = &ipc->fd;
-	bfd->fd = ret;
-	bfd->data = ipc;
-	bfd->priv_nr = listen_bfd->priv_nr;
-	bfd->cb = proxy_ipaccess_fd_cb;
-	bfd->when = BSC_FD_READ;
+	osmo_fd_setup(bfd, ret, OSMO_FD_READ, proxy_ipaccess_fd_cb, ipc, listen_bfd->priv_nr);
 	ret = osmo_fd_register(bfd);
 	if (ret < 0) {
 		LOGP(DLINP, LOGL_ERROR, "could not register FD\n");
@@ -1016,20 +1012,17 @@ static struct ipa_proxy_conn *connect_bsc(struct sockaddr_in *sa, int priv_nr, v
 
 	ipc->bts_conn = data;
 
-	bfd = &ipc->fd;
-	bfd->fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
-	bfd->cb = ipaccess_fd_cb;
-	bfd->when = BSC_FD_READ | BSC_FD_WRITE;
-	bfd->data = ipc;
-	bfd->priv_nr = priv_nr;
-
-	if (bfd->fd < 0) {
+	ret = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
+	if (ret < 0) {
 		LOGP(DLINP, LOGL_ERROR, "Could not create socket: %s\n",
 			strerror(errno));
 		talloc_free(ipc);
 		return NULL;
 	}
 
+	bfd = &ipc->fd;
+	osmo_fd_setup(bfd, ret, OSMO_FD_READ | OSMO_FD_WRITE, ipaccess_fd_cb, ipc, priv_nr);
+
 	ret = setsockopt(bfd->fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
 	if (ret < 0) {
 		LOGP(DLINP, LOGL_ERROR, "Could not set socket option\n");
@@ -1093,14 +1086,22 @@ static int ipaccess_proxy_setup(void)
 	return ret;
 }
 
-static void signal_handler(int signal)
+static void signal_handler(int signum)
 {
-	fprintf(stdout, "signal %u received\n", signal);
+	fprintf(stdout, "signal %u received\n", signum);
 
-	switch (signal) {
+	switch (signum) {
 	case SIGABRT:
-		/* in case of abort, we want to obtain a talloc report
-		 * and then return to the caller, who will abort the process */
+		/* in case of abort, we want to obtain a talloc report and
+		 * then run default SIGABRT handler, who will generate coredump
+		 * and abort the process. abort() should do this for us after we
+		 * return, but program wouldn't exit if an external SIGABRT is
+		 * received.
+		 */
+		talloc_report_full(tall_bsc_ctx, stderr);
+		signal(SIGABRT, SIG_DFL);
+		raise(SIGABRT);
+		break;
 	case SIGUSR1:
 		talloc_report_full(tall_bsc_ctx, stderr);
 		break;
@@ -1254,7 +1255,9 @@ int main(int argc, char **argv)
 }
 
 /* Stub */
-int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan)
-{
-	return 0;
-}
+int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan) { return 0; }
+void pcu_info_update(struct gsm_bts *bts) {};
+int rsl_sacch_filling(struct gsm_bts_trx *trx, uint8_t type, const uint8_t *data, int len) { return 0; }
+int rsl_bcch_info(const struct gsm_bts_trx *trx, enum osmo_sysinfo_type si_type, const uint8_t *data, int len)
+{ return 0; }
+int gsm_generate_si(struct gsm_bts *bts, enum osmo_sysinfo_type si_type) { return 0; }
