@@ -33,6 +33,7 @@
 #include <osmocom/bsc/debug.h>
 #include <osmocom/bsc/signal.h>
 #include <osmocom/bsc/abis_om2000.h>
+#include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/vty.h>
 
 #include <osmocom/vty/vty.h>
@@ -40,6 +41,8 @@
 #include <osmocom/vty/logging.h>
 #include <osmocom/vty/telnet_interface.h>
 
+#define X(x) (1 << x)
+
 static struct cmd_node om2k_node = {
 	OM2K_NODE,
 	"%s(om2k)# ",
@@ -66,13 +69,15 @@ static int dummy_config_write(struct vty *v)
 }
 
 /* FIXME: auto-generate those strings from the value_string lists */
-#define OM2K_OBJCLASS_VTY "(trxc|ts|tf|is|con|dp|cf|tx|rx)"
+#define OM2K_OBJCLASS_VTY "(trxc|tg|ts|tf|is|con|dp|mctr|cf|tx|rx)"
 #define OM2K_OBJCLASS_VTY_HELP 	"TRX Controller\n"	\
+				"Trunk Group\n"		\
 				"Timeslot\n"		\
 				"Timing Function\n"	\
 				"Interface Switch\n"	\
 				"Abis Concentrator\n"	\
 				"Digital Path\n"	\
+				"Multi Carrier Transceiver\n" \
 				"Central Function\n"	\
 				"Transmitter\n"		\
 				"Receiver\n"
@@ -253,6 +258,34 @@ DEFUN(om2k_cap_req, om2k_cap_req_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(om2k_arb, om2k_arb_cmd,
+	"arbitrary <0-65535> [HEXSTRING]",
+	"Send arbitrary OM2k message\n"
+	"Command identifier\n"
+	"Hex Encoded payload\n")
+{
+	struct oml_node_state *oms = vty->index;
+	uint8_t buf[128];
+	int rc;
+	int cmd = atoi(argv[0]);
+
+	if (argc >= 2) {
+		rc = osmo_hexparse(argv[1], buf, sizeof(buf));
+		if (rc < 0 || rc > sizeof(buf)) {
+			vty_out(vty, "Error parsing HEXSTRING%s", VTY_NEWLINE);
+			return CMD_WARNING;
+		}
+	} else {
+		rc = 0;
+	}
+
+	abis_om2k_tx_arb(oms->bts, &oms->mo, cmd, buf, rc);
+
+	return CMD_SUCCESS;
+}
+
+
+
 static struct con_group *con_group_find_or_create(struct gsm_bts *bts, uint8_t cg)
 {
 	struct con_group *ent;
@@ -312,10 +345,11 @@ static int con_group_del_path(struct con_group *cg, uint16_t ccp,
 	return -ENOENT;
 }
 
-DEFUN(cfg_om2k_con_group, cfg_om2k_con_group_cmd,
-	"con-connection-group <1-31>",
-	"Configure a CON (Concentrator) Connection Group\n"
-	"CON Connection Group Number\n")
+DEFUN_ATTR(cfg_om2k_con_group, cfg_om2k_con_group_cmd,
+	   "con-connection-group <1-31>",
+	   "Configure a CON (Concentrator) Connection Group\n"
+	   "CON Connection Group Number\n",
+	   CMD_ATTR_IMMEDIATE)
 {
 	struct gsm_bts *bts = vty->index;
 	struct con_group *cg;
@@ -340,10 +374,11 @@ DEFUN(cfg_om2k_con_group, cfg_om2k_con_group_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(del_om2k_con_group, del_om2k_con_group_cmd,
-	"del-connection-group <1-31>",
-	"Delete a CON (Concentrator) Connection Group\n"
-	"CON Connection Group Number\n")
+DEFUN_ATTR(del_om2k_con_group, del_om2k_con_group_cmd,
+	   "del-connection-group <1-31>",
+	   "Delete a CON (Concentrator) Connection Group\n"
+	   "CON Connection Group Number\n",
+	   CMD_ATTR_IMMEDIATE)
 {
 	struct gsm_bts *bts = vty->index;
 	int rc;
@@ -371,9 +406,10 @@ DEFUN(del_om2k_con_group, del_om2k_con_group_cmd,
 			"CON Connection Point\n"				\
 			"Contiguity Index\n"				\
 
-DEFUN(cfg_om2k_con_path_dec, cfg_om2k_con_path_dec_cmd,
-	"con-path (add|del) <0-2047> <0-255> deconcentrated <0-63>",
-	CON_PATH_HELP "De-concentrated in/outlet\n" "TEI Value\n")
+DEFUN_USRATTR(cfg_om2k_con_path_dec, cfg_om2k_con_path_dec_cmd,
+	      X(BSC_VTY_ATTR_RESTART_ABIS_OML_LINK),
+	      "con-path (add|del) <0-2047> <0-255> deconcentrated <0-63>",
+	      CON_PATH_HELP "De-concentrated in/outlet\n" "TEI Value\n")
 {
 	struct con_group *cg = vty->index;
 	uint16_t ccp = atoi(argv[1]);
@@ -393,9 +429,10 @@ DEFUN(cfg_om2k_con_path_dec, cfg_om2k_con_path_dec_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_om2k_con_path_conc, cfg_om2k_con_path_conc_cmd,
-	"con-path (add|del) <0-2047> <0-255> concentrated <1-16>",
-	CON_PATH_HELP "Concentrated in/outlet\n" "Tag Number\n")
+DEFUN_USRATTR(cfg_om2k_con_path_conc, cfg_om2k_con_path_conc_cmd,
+	      X(BSC_VTY_ATTR_RESTART_ABIS_OML_LINK),
+	      "con-path (add|del) <0-2047> <0-255> concentrated <1-16>",
+	      CON_PATH_HELP "Concentrated in/outlet\n" "Tag Number\n")
 {
 	struct con_group *cg = vty->index;
 	uint16_t ccp = atoi(argv[1]);
@@ -415,11 +452,12 @@ DEFUN(cfg_om2k_con_path_conc, cfg_om2k_con_path_conc_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_bts_alt_mode, cfg_bts_alt_mode_cmd,
-	"abis-lower-transport (single-timeslot|super-channel)",
-	"Configure thee Abis Lower Transport\n"
-	"Single Timeslot (classic Abis)\n"
-	"SuperChannel (Packet Abis)\n")
+DEFUN_USRATTR(cfg_bts_alt_mode, cfg_bts_alt_mode_cmd,
+	      X(BSC_VTY_ATTR_RESTART_ABIS_OML_LINK),
+	      "abis-lower-transport (single-timeslot|super-channel)",
+	      "Configure thee Abis Lower Transport\n"
+	      "Single Timeslot (classic Abis)\n"
+	      "SuperChannel (Packet Abis)\n")
 {
 	struct gsm_bts *bts = vty->index;
 
@@ -437,11 +475,47 @@ DEFUN(cfg_bts_alt_mode, cfg_bts_alt_mode_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(cfg_bts_is_conn_list, cfg_bts_is_conn_list_cmd,
-	"is-connection-list (add|del) <0-2047> <0-2047> <0-255>",
-	"Interface Switch Connection List\n"
-	"Add to IS list\n" "Delete from IS list\n"
-	"ICP1\n" "ICP2\n" "Contiguity Index\n")
+DEFUN_USRATTR(cfg_bts_om2k_version_limit, cfg_bts_om2k_version_limit_cmd,
+	      X(BSC_VTY_ATTR_RESTART_ABIS_OML_LINK),
+	      "om2000 version-limit (oml|rsl) gen <0-99> rev <0-99>",
+	      "Configure OM2K specific parameters\n"
+	      "Configure optional maximum protocol version to negotiate\n"
+	      "Limit OML IWD version\n" "Limit RSL IWD version\n"
+	      "Generation limit\n"
+	      "Generation number to limit to (inclusive)\n"
+	      "Revision limit\n"
+	      "Revision number to limit to (inclusive)\n")
+{
+	struct gsm_bts *bts = vty->index;
+	int iwd;
+
+	if (bts->type != GSM_BTS_TYPE_RBS2000) {
+		vty_out(vty, "%% Command only works for RBS2000%s",
+			VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (!strcmp(argv[0], "oml"))
+		iwd = 0;
+	else if (!strcmp(argv[0], "rsl"))
+		iwd = 1;
+	else {
+		vty_out(vty, "%% Invalid IWD%s",
+			VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	bts->rbs2000.om2k_version[iwd].limit = (atoi(argv[1]) << 8) | atoi(argv[2]);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN_USRATTR(cfg_bts_is_conn_list, cfg_bts_is_conn_list_cmd,
+	      X(BSC_VTY_ATTR_RESTART_ABIS_OML_LINK),
+	      "is-connection-list (add|del) <0-2047> <0-2047> <0-255>",
+	      "Interface Switch Connection List\n"
+	      "Add to IS list\n" "Delete from IS list\n"
+	      "ICP1\n" "ICP2\n" "Contiguity Index\n")
 {
 	struct gsm_bts *bts = vty->index;
 	uint16_t icp1 = atoi(argv[1]);
@@ -492,8 +566,11 @@ DEFUN(om2k_conf_req, om2k_conf_req_cmd,
 	case OM2K_MO_CLS_IS:
 		abis_om2k_tx_is_conf_req(bts);
 		break;
+	case OM2K_MO_CLS_CON:
+		abis_om2k_tx_con_conf_req(bts);
+		break;
 	case OM2K_MO_CLS_TS:
-		trx = gsm_bts_trx_by_nr(bts, oms->mo.assoc_so);
+		trx = gsm_bts_trx_num(bts, oms->mo.assoc_so);
 		if (!trx) {
 			vty_out(vty, "%% BTS %u has no TRX %u%s", bts->nr,
 				oms->mo.assoc_so, VTY_NEWLINE);
@@ -510,7 +587,7 @@ DEFUN(om2k_conf_req, om2k_conf_req_cmd,
 	case OM2K_MO_CLS_RX:
 	case OM2K_MO_CLS_TX:
 	case OM2K_MO_CLS_TRXC:
-		trx = gsm_bts_trx_by_nr(bts, oms->mo.inst);
+		trx = gsm_bts_trx_num(bts, oms->mo.inst);
 		if (!trx) {
 			vty_out(vty, "%% BTS %u has no TRX %u%s", bts->nr,
 				oms->mo.inst, VTY_NEWLINE);
@@ -558,6 +635,7 @@ void abis_om2k_config_write_bts(struct vty *vty, struct gsm_bts *bts)
 {
 	struct is_conn_group *igrp;
 	struct con_group *cgrp;
+	unsigned int i;
 
 	llist_for_each_entry(igrp, &bts->rbs2000.is.conn_groups, list)
 		vty_out(vty, "  is-connection-list add %u %u %u%s",
@@ -571,10 +649,89 @@ void abis_om2k_config_write_bts(struct vty *vty, struct gsm_bts *bts)
 	if (bts->rbs2000.use_superchannel)
 		vty_out(vty, "  abis-lower-transport super-channel%s",
 			VTY_NEWLINE);
+	for (i = 0; i < 2; i++)
+		if (bts->rbs2000.om2k_version[i].limit)
+			vty_out(vty, "  om2000 version-limit %s gen %02d rev %02d%s",
+				i ? "rsl" : "oml",
+				(bts->rbs2000.om2k_version[i].limit >> 8),
+				(bts->rbs2000.om2k_version[i].limit & 0xff),
+				VTY_NEWLINE);
+}
+
+static void vty_dump_om2k_mo(struct vty *vty, const struct om2k_mo *mo, const char *pfx)
+{
+	unsigned int pfx_len = strlen(pfx);
+	const char *mo_name = abis_om2k_mo_name(&mo->addr);
+	unsigned int pfx_mo_len = pfx_len + strlen(mo_name);
+	unsigned int pfx2_len;
+	char pfx2[23];
+	int i;
+
+	/* generate padding after MO class to align the state names in the same column */
+	if (pfx_mo_len > sizeof(pfx2)-1)
+		pfx2_len = 0;
+	else
+		pfx2_len = sizeof(pfx2)-1 - pfx_mo_len;
+	for (i = 0; i < pfx2_len; i++)
+		pfx2[i] = ' ';
+	pfx2[pfx2_len] = '\0';
+
+	vty_out(vty, "%s%s%s %s%s", pfx, mo_name, pfx2,
+		mo->fsm ? osmo_fsm_inst_state_name(mo->fsm) : "[NULL]",
+		VTY_NEWLINE);
+}
+
+DEFUN(show_om2k_mo, show_om2k_mo_cmd,
+	"show bts <0-255> om2k-mo",
+	SHOW_STR "Display information about a BTS\n"
+	"BTS number\n" "OM2000 Managed Object information\n")
+{
+	struct gsm_network *net = gsmnet_from_vty(vty);
+	int bts_nr = atoi(argv[0]);
+	struct gsm_bts *bts = gsm_bts_num(net, bts_nr);
+	struct gsm_bts_trx *trx;
+
+	if (!bts) {
+		vty_out(vty, "%% can't find BTS '%s'%s", argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (bts->type != GSM_BTS_TYPE_RBS2000) {
+		vty_out(vty, "%% BTS is not using OM2000%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	vty_out(vty, "BTS %3u OM2K-FSM state %s%s", bts->nr,
+		osmo_fsm_inst_state_name(bts->rbs2000.bts_fi), VTY_NEWLINE);
+	vty_dump_om2k_mo(vty, &bts->rbs2000.cf.om2k_mo, " ");
+	vty_dump_om2k_mo(vty, &bts->rbs2000.con.om2k_mo, " ");
+	vty_dump_om2k_mo(vty, &bts->rbs2000.is.om2k_mo, " ");
+	vty_dump_om2k_mo(vty, &bts->rbs2000.dp.om2k_mo, " ");
+	vty_dump_om2k_mo(vty, &bts->rbs2000.tf.om2k_mo, " ");
+	vty_dump_om2k_mo(vty, &bts->rbs2000.mctr.om2k_mo, " ");
+
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		int tn;
+
+		vty_out(vty, " TRX %u OM2K-FSM state  %s%s", trx->nr,
+			osmo_fsm_inst_state_name(trx->rbs2000.trx_fi), VTY_NEWLINE);
+		vty_dump_om2k_mo(vty, &trx->rbs2000.trxc.om2k_mo, "  ");
+		vty_dump_om2k_mo(vty, &trx->rbs2000.rx.om2k_mo, "  ");
+		vty_dump_om2k_mo(vty, &trx->rbs2000.tx.om2k_mo, "  ");
+
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+			vty_dump_om2k_mo(vty, &ts->rbs2000.om2k_mo, "   ");
+		}
+	}
+
+	return CMD_SUCCESS;
 }
 
 int abis_om2k_vty_init(void)
 {
+	install_element_ve(&show_om2k_mo_cmd);
 	install_element(ENABLE_NODE, &om2k_class_inst_cmd);
 	install_element(ENABLE_NODE, &om2k_classnum_inst_cmd);
 	install_node(&om2k_node, dummy_config_write);
@@ -590,6 +747,7 @@ int abis_om2k_vty_init(void)
 	install_element(OM2K_NODE, &om2k_test_cmd);
 	install_element(OM2K_NODE, &om2k_cap_req_cmd);
 	install_element(OM2K_NODE, &om2k_conf_req_cmd);
+	install_element(OM2K_NODE, &om2k_arb_cmd);
 
 	install_node(&om2k_con_group_node, dummy_config_write);
 	install_element(OM2K_CON_GROUP_NODE, &cfg_om2k_con_path_dec_cmd);
@@ -597,6 +755,7 @@ int abis_om2k_vty_init(void)
 
 	install_element(BTS_NODE, &cfg_bts_is_conn_list_cmd);
 	install_element(BTS_NODE, &cfg_bts_alt_mode_cmd);
+	install_element(BTS_NODE, &cfg_bts_om2k_version_limit_cmd);
 	install_element(BTS_NODE, &cfg_om2k_con_group_cmd);
 	install_element(BTS_NODE, &del_om2k_con_group_cmd);
 
