@@ -32,8 +32,10 @@
 #include <osmocom/bsc/osmo_bsc_lcls.h>
 #include <osmocom/bsc/bsc_msc_data.h>
 #include <osmocom/bsc/gsm_08_08.h>
+#include <osmocom/bsc/gsm_04_08_rr.h>
 #include <osmocom/bsc/lchan_select.h>
 #include <osmocom/bsc/abis_rsl.h>
+#include <osmocom/bsc/bts.h>
 
 #include <osmocom/bsc/assignment_fsm.h>
 
@@ -69,16 +71,34 @@ static const struct osmo_tdef_state_timeout assignment_fsm_timeouts[32] = {
 		_conn->assignment.failure_cause = cause; \
 		LOG_ASSIGNMENT(_conn, LOGL_ERROR, "Assignment failed in state %s, cause %s: " fmt "\n", \
 			       osmo_fsm_inst_state_name(fi), gsm0808_cause_name(cause), ## args); \
-		assignment_count_result(BSC_CTR_ASSIGNMENT_ERROR); \
+		assignment_count_result(CTR_ASSIGNMENT_ERROR); \
 		on_assignment_failure(_conn); \
 	} while(0)
 
 /* Assume presence of local var 'conn' as struct gsm_subscriber_connection */
 #define assignment_count(counter) do { \
+		struct gsm_bts *bts = conn_get_bts(conn); \
 		LOG_ASSIGNMENT(conn, LOGL_DEBUG, "incrementing rate counter: %s %s\n", \
-			       bsc_ctr_description[counter].name, \
-			       bsc_ctr_description[counter].description); \
-		rate_ctr_inc(&conn->network->bsc_ctrs->ctr[counter]); \
+			       bsc_ctr_description[BSC_##counter].name, \
+			       bsc_ctr_description[BSC_##counter].description); \
+		rate_ctr_inc(&conn->network->bsc_ctrs->ctr[BSC_##counter]); \
+		if (bts) { \
+			rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_##counter]); \
+			if (BTS_##counter != BTS_CTR_ASSIGNMENT_NO_CHANNEL) { \
+				switch (conn->lchan->ch_mode_rate.chan_mode) { \
+				case GSM48_CMODE_SIGN: \
+					rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_##counter##_SIGN]); \
+					break; \
+				case GSM48_CMODE_SPEECH_V1: \
+				case GSM48_CMODE_SPEECH_EFR: \
+				case GSM48_CMODE_SPEECH_AMR: \
+					rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_##counter##_SPEECH]); \
+					break; \
+				default: \
+					break; \
+				} \
+			} \
+		} \
 	} while(0)
 
 #define assignment_count_result(counter) do { \
@@ -88,8 +108,8 @@ static const struct osmo_tdef_state_timeout assignment_fsm_timeouts[32] = {
 		} else \
 			LOG_ASSIGNMENT(conn, LOGL_DEBUG, \
 				       "result rate counter already recorded, NOT counting as: %s %s\n", \
-				       bsc_ctr_description[counter].name, \
-				       bsc_ctr_description[counter].description); \
+				       bsc_ctr_description[BSC_##counter].name, \
+				       bsc_ctr_description[BSC_##counter].description); \
 	} while(0)
 
 void assignment_reset(struct gsm_subscriber_connection *conn)
@@ -116,10 +136,12 @@ static void on_assignment_failure(struct gsm_subscriber_connection *conn)
 {
 	struct msgb *resp = gsm0808_create_assignment_failure(conn->assignment.failure_cause, NULL);
 
-	if (!resp)
+	if (!resp) {
 		LOG_ASSIGNMENT(conn, LOGL_ERROR, "Unable to compose BSSMAP Assignment Failure message\n");
-	else
+	} else {
+		rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_ASSIGMENT_FAILURE]);
 		gscon_sigtran_send(conn, resp);
+	}
 
 	/* If assignment failed as early as in assignment_fsm_start(), there may not be an fi yet. */
 	if (conn->assignment.fi) {
@@ -206,6 +228,7 @@ static void send_assignment_complete(struct gsm_subscriber_connection *conn)
 	    conn->assignment.req.use_osmux)
 		_gsm0808_ass_compl_extend_osmux(resp, osmux_cid);
 
+	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_ASSIGMENT_COMPLETE]);
 	rc = gscon_sigtran_send(conn, resp);
 	if (rc) {
 		assignment_fail(GSM0808_CAUSE_EQUIPMENT_FAILURE,
@@ -217,10 +240,19 @@ static void send_assignment_complete(struct gsm_subscriber_connection *conn)
 
 static void assignment_success(struct gsm_subscriber_connection *conn)
 {
+	struct gsm_bts *bts;
+
 	/* Take on the new lchan */
 	gscon_change_primary_lchan(conn, conn->assignment.new_lchan);
 	conn->assignment.new_lchan = NULL;
 
+	OSMO_ASSERT((bts = conn_get_bts(conn)) != NULL);
+	if (is_siemens_bts(bts) && ts_is_tch(conn->lchan->ts)) {
+		/* HACK: store the actual Classmark 2 LV from the subscriber and use it here! */
+		uint8_t cm2_lv[] = { 0x02, 0x00, 0x00 };
+		send_siemens_mrpci(conn->lchan, cm2_lv);
+	}
+
 	/* apply LCLS configuration (if any) */
 	lcls_apply_config(conn);
 
@@ -247,7 +279,7 @@ static void assignment_success(struct gsm_subscriber_connection *conn)
 	LOG_ASSIGNMENT(conn, LOGL_DEBUG, "Assignment successful\n");
 	osmo_fsm_inst_term(conn->assignment.fi, OSMO_FSM_TERM_REGULAR, 0);
 
-	assignment_count_result(BSC_CTR_ASSIGNMENT_COMPLETED);
+	assignment_count_result(CTR_ASSIGNMENT_COMPLETED);
 }
 
 static void assignment_fsm_update_id(struct gsm_subscriber_connection *conn)
@@ -372,9 +404,8 @@ static int check_requires_voice_stream(struct gsm_subscriber_connection *conn)
 	return 0;
 }
 
-/* Check if the conn is already associated with an lchan. If yes, we will check
- * if that lchan is compatible with the preferred rate/codec. If the lchan
- * turns out to be incompatible we try with the alternate rate/codec. */
+/* Decide if we should re-use an existing lchan. For this we check if the
+ * current lchan is compatible with one of the requested modes. */
 static bool reuse_existing_lchan(struct gsm_subscriber_connection *conn)
 {
 	struct assignment_request *req = &conn->assignment.req;
@@ -388,22 +419,10 @@ static bool reuse_existing_lchan(struct gsm_subscriber_connection *conn)
 	for (i = 0; i < req->n_ch_mode_rate; i++)
 		if (lchan_type_compat_with_mode(conn->lchan->type, &req->ch_mode_rate[i])) {
 			conn->lchan->ch_mode_rate = req->ch_mode_rate[i];
-			break;
+			return true;
 		}
 
-	if (i == req->n_ch_mode_rate)
-		return false;
-
-	if (conn->lchan->tch_mode != conn->lchan->ch_mode_rate.chan_mode) {
-		/* FIXME: send Channel Mode Modify to put the current lchan in the right mode, and kick
-		 * off its RTP stream setup code path. See gsm48_lchan_modify() and
-		 * gsm48_rx_rr_modif_ack(), and see lchan_fsm.h LCHAN_EV_CHAN_MODE_MODIF_* */
-		LOG_ASSIGNMENT(conn, LOGL_DEBUG,
-			       "Current lchan would be compatible, but Channel Mode Modify is not implemented\n");
-		return false;
-	}
-
-	return true;
+	return false;
 }
 
 void assignment_fsm_start(struct gsm_subscriber_connection *conn, struct gsm_bts *bts,
@@ -423,7 +442,7 @@ void assignment_fsm_start(struct gsm_subscriber_connection *conn, struct gsm_bts
 	OSMO_ASSERT(!conn->assignment.fi);
 	OSMO_ASSERT(!conn->assignment.new_lchan);
 
-	assignment_count(BSC_CTR_ASSIGNMENT_ATTEMPTED);
+	assignment_count(CTR_ASSIGNMENT_ATTEMPTED);
 
 	fi = osmo_fsm_inst_alloc_child(&assignment_fsm, conn->fi, GSCON_EV_ASSIGNMENT_END);
 	OSMO_ASSERT(fi);
@@ -440,22 +459,62 @@ void assignment_fsm_start(struct gsm_subscriber_connection *conn, struct gsm_bts
 		return;
 
 	/* There may be an already existing lchan, if yes, try to work with
-	 * the existing lchan */
-	if (reuse_existing_lchan(conn)) {
+	 * the existing lchan.
+	 * If an RTP FSM is already set up for the lchan, Mode Modify is not yet supported -- see handling of
+	 * LCHAN_EV_REQUEST_MODE_MODIFY in lchan_fsm.c. To not break the lchan, do not even attempt to re-use an lchan
+	 * that already has an RTP stream set up, rather establish a new lchan (that transition is well implemented). */
+	if (reuse_existing_lchan(conn) && !conn->lchan->fi_rtp) {
+
+		/* If the requested mode and the current TCH mode matches up, just send the
+		 * assignment complete directly and be done with the assignment procedure. */
+		if (conn->lchan->tch_mode == conn->lchan->ch_mode_rate.chan_mode) {
+			LOG_ASSIGNMENT(conn, LOGL_DEBUG,
+				       "Current lchan mode is compatible with requested chan_mode,"
+				       " sending BSSMAP Assignment Complete directly."
+				       " requested chan_mode=%s; current lchan is %s\n",
+				       gsm48_chan_mode_name(conn->lchan->ch_mode_rate.chan_mode),
+				       gsm_lchan_name(conn->lchan));
+
+			send_assignment_complete(conn);
+			/* If something went wrong during send_assignment_complete(),
+			 * the fi will be gone from error handling in there. */
+			if (conn->assignment.fi) {
+				assignment_count_result(CTR_ASSIGNMENT_COMPLETED);
+				osmo_fsm_inst_term(conn->assignment.fi, OSMO_FSM_TERM_REGULAR, 0);
+			}
+			return;
+		}
+
+		/* The requested mode does not match the current TCH mode but the lchan is
+		 * compatible. We will initiate a mode modify procedure. */
 		LOG_ASSIGNMENT(conn, LOGL_DEBUG,
-			       "Current lchan is compatible with requested chan_mode,"
-			       " sending BSSMAP Assignment Complete directly."
-			       " requested chan_mode=%s; current lchan is %s\n",
+			       "Current lchan mode is not compatible with requested chan_mode,"
+			       " so we will modify it. requested chan_mode=%s; current lchan is %s\n",
 			       gsm48_chan_mode_name(conn->lchan->ch_mode_rate.chan_mode),
 			       gsm_lchan_name(conn->lchan));
 
-		send_assignment_complete(conn);
-		/* If something went wrong during send_assignment_complete(), the fi will be gone from
-		 * error handling in there. */
-		if (conn->assignment.fi) {
-			assignment_count_result(BSC_CTR_ASSIGNMENT_COMPLETED);
-			osmo_fsm_inst_term(conn->assignment.fi, OSMO_FSM_TERM_REGULAR, 0);
-		}
+		info = (struct lchan_activate_info){
+			.activ_for = FOR_ASSIGNMENT,
+			.for_conn = conn,
+			.chan_mode = conn->lchan->ch_mode_rate.chan_mode,
+			.encr = conn->lchan->encr,
+			.s15_s0 = conn->lchan->ch_mode_rate.s15_s0,
+			.requires_voice_stream = conn->assignment.requires_voice_stream,
+			.msc_assigned_cic = req->msc_assigned_cic,
+			.re_use_mgw_endpoint_from_lchan = conn->lchan,
+			.ta = conn->lchan->last_ta,
+			.ta_known = true,
+		};
+
+		osmo_fsm_inst_dispatch(conn->lchan->fi, LCHAN_EV_REQUEST_MODE_MODIFY, &info);
+
+		/* Since we opted not to allocate a new lchan, the new lchan is still the old lchan. */
+		conn->assignment.new_lchan = conn->lchan;
+
+		/* Also we need to skip the RR assignment, so we jump forward and wait for the lchan_fsm until it
+		 * reaches the established state again. */
+		assignment_fsm_state_chg(ASSIGNMENT_ST_WAIT_LCHAN_ESTABLISHED);
+
 		return;
 	}
 
@@ -463,6 +522,11 @@ void assignment_fsm_start(struct gsm_subscriber_connection *conn, struct gsm_bts
 	for (i = 0; i < req->n_ch_mode_rate; i++) {
 		conn->assignment.new_lchan = lchan_select_by_chan_mode(bts,
 		    req->ch_mode_rate[i].chan_mode, req->ch_mode_rate[i].chan_rate);
+		/* FIXME: at this point there is merely an assignment request with a given ch_mode_rate. Writing this to
+		 * conn->lchan->ch_mode_rate is a violation of scopes: the lchan->* state should only be modified
+		 * *after* the assignment is confirmed to be completed. Before that, this data should live in
+		 * conn->assignment or the lchan_activate_info, the designated places for not-yet-confirmed data. See
+		 * OS#3833 */
 		conn->lchan->ch_mode_rate = req->ch_mode_rate[i];
 		if (conn->assignment.new_lchan)
 			break;
@@ -471,7 +535,19 @@ void assignment_fsm_start(struct gsm_subscriber_connection *conn, struct gsm_bts
 	/* Check whether the lchan allocation was successful or not and tear
 	 * down the assignment in case of failure. */
 	if (!conn->assignment.new_lchan) {
-		assignment_count_result(BSC_CTR_ASSIGNMENT_NO_CHANNEL);
+		assignment_count_result(CTR_ASSIGNMENT_NO_CHANNEL);
+		switch (req->ch_mode_rate[0].chan_mode) {
+		case GSM48_CMODE_SIGN:
+			rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_ASSIGNMENT_NO_CHANNEL_SIGN]);
+			break;
+		case GSM48_CMODE_SPEECH_V1:
+		case GSM48_CMODE_SPEECH_EFR:
+		case GSM48_CMODE_SPEECH_AMR:
+			rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_ASSIGNMENT_NO_CHANNEL_SPEECH]);
+			break;
+		default:
+			break;
+		}
 		assignment_fail(GSM0808_CAUSE_NO_RADIO_RESOURCE_AVAILABLE,
 				"BSSMAP Assignment Command:"
 				" No lchan available for: pref=%s:%s / alt1=%s:%s / alt2=%s:%s\n",
@@ -503,6 +579,8 @@ void assignment_fsm_start(struct gsm_subscriber_connection *conn, struct gsm_bts
 		.requires_voice_stream = conn->assignment.requires_voice_stream,
 		.msc_assigned_cic = req->msc_assigned_cic,
 		.re_use_mgw_endpoint_from_lchan = conn->lchan,
+		.ta = conn->lchan->last_ta,
+		.ta_known = true,
 	};
 	lchan_activate(conn->assignment.new_lchan, &info);
 }
@@ -562,7 +640,7 @@ static void assignment_fsm_wait_rr_ass_complete(struct osmo_fsm_inst *fi, uint32
 		return;
 
 	case ASSIGNMENT_EV_RR_ASSIGNMENT_FAIL:
-		assignment_count_result(BSC_CTR_ASSIGNMENT_FAILED);
+		assignment_count_result(CTR_ASSIGNMENT_FAILED);
 		assignment_fail(get_cause(data), "Rx RR Assignment Failure");
 		return;
 
@@ -675,6 +753,7 @@ static const struct osmo_fsm_state assignment_fsm_states[] = {
 		.out_state_mask = 0
 			| S(ASSIGNMENT_ST_WAIT_LCHAN_ACTIVE)
 			| S(ASSIGNMENT_ST_WAIT_RR_ASS_COMPLETE)
+			| S(ASSIGNMENT_ST_WAIT_LCHAN_ESTABLISHED) /* MODE MODIFY */
 			,
 	},
 	[ASSIGNMENT_ST_WAIT_RR_ASS_COMPLETE] = {
@@ -730,7 +809,7 @@ void assignment_fsm_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, vo
 	switch (event) {
 
 	case ASSIGNMENT_EV_CONN_RELEASING:
-		assignment_count_result(BSC_CTR_ASSIGNMENT_STOPPED);
+		assignment_count_result(CTR_ASSIGNMENT_STOPPED);
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REQUEST, 0);
 		return;
 
@@ -750,7 +829,7 @@ void assignment_fsm_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, vo
 int assignment_fsm_timer_cb(struct osmo_fsm_inst *fi)
 {
 	struct gsm_subscriber_connection *conn = assignment_fi_conn(fi);
-	assignment_count_result(BSC_CTR_ASSIGNMENT_TIMEOUT);
+	assignment_count_result(CTR_ASSIGNMENT_TIMEOUT);
 	assignment_fail(GSM0808_CAUSE_EQUIPMENT_FAILURE, "Timeout");
 	return 0;
 }
