@@ -23,7 +23,8 @@
 #include <osmocom/core/msgb.h>
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/abis_nm.h>
-
+#include <osmocom/bsc/bts.h>
+#include <osmocom/gsm/bts_features.h>
 
 struct msgb *nanobts_attr_bts_get(struct gsm_bts *bts)
 {
@@ -31,6 +32,8 @@ struct msgb *nanobts_attr_bts_get(struct gsm_bts *bts)
 	uint8_t buf[256];
 	int rlt;
 	msgb = msgb_alloc(1024, "nanobts_attr_bts");
+	if (!msgb)
+		return NULL;
 
 	memcpy(buf, "\x55\x5b\x61\x67\x6d\x73", 6);
 	msgb_tv_fixed_put(msgb, NM_ATT_INTERF_BOUND, 6, buf);
@@ -98,20 +101,23 @@ struct msgb *nanobts_attr_bts_get(struct gsm_bts *bts)
 	return msgb;
 }
 
-struct msgb *nanobts_attr_nse_get(struct gsm_bts *bts)
+struct msgb *nanobts_attr_nse_get(struct gsm_bts_sm *bts_sm)
 {
 	struct msgb *msgb;
 	uint8_t buf[256];
+	struct gsm_bts *bts = gsm_bts_sm_get_bts(bts_sm);
 	msgb = msgb_alloc(1024, "nanobts_attr_bts");
+	if (!msgb)
+		return NULL;
 
 	/* NSEI 925 */
-	buf[0] = bts->gprs.nse.nsei >> 8;
-	buf[1] = bts->gprs.nse.nsei & 0xff;
+	buf[0] = bts_sm->gprs.nse.nsei >> 8;
+	buf[1] = bts_sm->gprs.nse.nsei & 0xff;
 	msgb_tl16v_put(msgb, NM_ATT_IPACC_NSEI, 2, buf);
 
 	/* all timers in seconds */
-	OSMO_ASSERT(ARRAY_SIZE(bts->gprs.nse.timer) < sizeof(buf));
-	memcpy(buf, bts->gprs.nse.timer, ARRAY_SIZE(bts->gprs.nse.timer));
+	OSMO_ASSERT(ARRAY_SIZE(bts_sm->gprs.nse.timer) < sizeof(buf));
+	memcpy(buf, bts_sm->gprs.nse.timer, ARRAY_SIZE(bts_sm->gprs.nse.timer));
 	msgb_tl16v_put(msgb, NM_ATT_IPACC_NS_CFG, 7, buf);
 
 	/* all timers in seconds */
@@ -139,6 +145,8 @@ struct msgb *nanobts_attr_cell_get(struct gsm_bts *bts)
 	struct msgb *msgb;
 	uint8_t buf[256];
 	msgb = msgb_alloc(1024, "nanobts_attr_bts");
+	if (!msgb)
+		return NULL;
 
 	/* routing area code */
 	buf[0] = bts->gprs.rac;
@@ -191,24 +199,47 @@ struct msgb *nanobts_attr_cell_get(struct gsm_bts *bts)
 	return msgb;
 }
 
-struct msgb *nanobts_attr_nscv_get(struct gsm_bts *bts)
+struct msgb *nanobts_attr_nsvc_get(struct gsm_bts *bts)
 {
 	struct msgb *msgb;
 	uint8_t buf[256];
+	struct gsm_bts_sm *bts_sm = bts->site_mgr;
 	msgb = msgb_alloc(1024, "nanobts_attr_bts");
+	if (!msgb)
+		return NULL;
 
 	/* 925 */
-	buf[0] = bts->gprs.nsvc[0].nsvci >> 8;
-	buf[1] = bts->gprs.nsvc[0].nsvci & 0xff;
+	buf[0] = bts_sm->gprs.nsvc[0].nsvci >> 8;
+	buf[1] = bts_sm->gprs.nsvc[0].nsvci & 0xff;
 	msgb_tl16v_put(msgb, NM_ATT_IPACC_NSVCI, 2, buf);
 
-	/* remote udp port */
-	osmo_store16be(bts->gprs.nsvc[0].remote_port, &buf[0]);
-	/* remote ip address */
-	osmo_store32be(bts->gprs.nsvc[0].remote_ip, &buf[2]);
-	/* local udp port */
-	osmo_store16be(bts->gprs.nsvc[0].local_port, &buf[6]);
-	msgb_tl16v_put(msgb, NM_ATT_IPACC_NS_LINK_CFG, 8, buf);
+	switch (bts_sm->gprs.nsvc->remote.u.sa.sa_family) {
+	case AF_INET6:
+		/* all fields are encoded in network byte order */
+		/* protocol family */
+		buf[0] = OSMO_NSVC_ADDR_IPV6;
+		/* padding */
+		buf[1] = 0x00;
+		/* local udp port */
+		osmo_store16be(bts_sm->gprs.nsvc[0].local_port, &buf[2]);
+		/* remote udp port */
+		memcpy(&buf[4], &bts_sm->gprs.nsvc[0].remote.u.sin6.sin6_port, sizeof(uint16_t));
+		/* remote ip address */
+		memcpy(&buf[6], &bts_sm->gprs.nsvc[0].remote.u.sin6.sin6_addr, sizeof(struct in6_addr));
+		msgb_tl16v_put(msgb, NM_ATT_OSMO_NS_LINK_CFG, 6 + sizeof(struct in6_addr), buf);
+		break;
+	case AF_INET:
+		/* remote udp port */
+		memcpy(&buf[0], &bts_sm->gprs.nsvc[0].remote.u.sin.sin_port, sizeof(uint16_t));
+		/* remote ip address */
+		memcpy(&buf[2], &bts_sm->gprs.nsvc[0].remote.u.sin.sin_addr, sizeof(struct in_addr));
+		/* local udp port */
+		osmo_store16be(bts_sm->gprs.nsvc[0].local_port, &buf[6]);
+		msgb_tl16v_put(msgb, NM_ATT_IPACC_NS_LINK_CFG, 8, buf);
+		break;
+	default:
+		break;
+	}
 
 	return msgb;
 }
@@ -219,6 +250,8 @@ struct msgb *nanobts_attr_radio_get(struct gsm_bts *bts,
 	struct msgb *msgb;
 	uint8_t buf[256];
 	msgb = msgb_alloc(1024, "nanobts_attr_bts");
+	if (!msgb)
+		return NULL;
 
 	/* number of -2dB reduction steps / Pn */
 	msgb_tv_put(msgb, NM_ATT_RF_MAXPOWR_R, trx->max_power_red / 2);
