@@ -39,7 +39,7 @@
 #include <osmocom/bsc/handover_fsm.h>
 #include <osmocom/bsc/bsc_msc_data.h>
 #include <osmocom/bsc/codec_pref.h>
-
+#include <osmocom/bsc/bts.h>
 
 static struct osmo_fsm lchan_fsm;
 
@@ -59,6 +59,7 @@ bool lchan_may_receive_data(struct gsm_lchan *lchan)
 	switch (lchan->fi->state) {
 	case LCHAN_ST_WAIT_RLL_RTP_ESTABLISH:
 	case LCHAN_ST_ESTABLISHED:
+	case LCHAN_ST_WAIT_RR_CHAN_MODE_MODIFY_ACK:
 		return true;
 	default:
 		return false;
@@ -209,13 +210,13 @@ static void lchan_on_fully_established(struct gsm_lchan *lchan)
 }
 
 struct osmo_tdef_state_timeout lchan_fsm_timeouts[32] = {
-	[LCHAN_ST_WAIT_TS_READY]	= { .T=23001 },
-	[LCHAN_ST_WAIT_ACTIV_ACK]	= { .T=23002 },
+	[LCHAN_ST_WAIT_TS_READY]	= { .T=-5 },
+	[LCHAN_ST_WAIT_ACTIV_ACK]	= { .T=-6 },
 	[LCHAN_ST_WAIT_RLL_RTP_ESTABLISH]	= { .T=3101 },
 	[LCHAN_ST_WAIT_RLL_RTP_RELEASED]	= { .T=3109 },
 	[LCHAN_ST_WAIT_BEFORE_RF_RELEASE]	= { .T=3111 },
 	[LCHAN_ST_WAIT_RF_RELEASE_ACK]	= { .T=3111 },
-	[LCHAN_ST_WAIT_AFTER_ERROR]	= { .T=993111 },
+	[LCHAN_ST_WAIT_AFTER_ERROR]	= { .T=-3111 },
 };
 
 /* Transition to a state, using the T timer defined in lchan_fsm_timeouts.
@@ -249,7 +250,7 @@ struct osmo_tdef_state_timeout lchan_fsm_timeouts[32] = {
 	} while(0)
 
 /* Which state to transition to when lchan_fail() is called in a given state. */
-uint32_t lchan_fsm_on_error[32] = {
+uint32_t lchan_fsm_on_error[34] = {
 	[LCHAN_ST_UNUSED] 			= LCHAN_ST_UNUSED,
 	[LCHAN_ST_WAIT_TS_READY] 		= LCHAN_ST_UNUSED,
 	[LCHAN_ST_WAIT_ACTIV_ACK] 		= LCHAN_ST_BORKEN,
@@ -260,6 +261,8 @@ uint32_t lchan_fsm_on_error[32] = {
 	[LCHAN_ST_WAIT_RF_RELEASE_ACK] 		= LCHAN_ST_BORKEN,
 	[LCHAN_ST_WAIT_AFTER_ERROR] 		= LCHAN_ST_UNUSED,
 	[LCHAN_ST_BORKEN] 			= LCHAN_ST_BORKEN,
+	[LCHAN_ST_WAIT_RR_CHAN_MODE_MODIFY_ACK]	= LCHAN_ST_BORKEN,
+	[LCHAN_ST_WAIT_RSL_CHAN_MODE_MODIFY_ACK] 	= LCHAN_ST_BORKEN,
 };
 
 #define lchan_fail(fmt, args...) lchan_fail_to(lchan_fsm_on_error[fi->state], fmt, ## args)
@@ -402,45 +405,88 @@ static void lchan_reset(struct gsm_lchan *lchan)
 		.meas_rep_last_seen_nr = 255,
 
 		.last_error = lchan->last_error,
+
+		.release.rr_cause = GSM48_RR_CAUSE_NORMAL,
 	};
 }
 
 static void lchan_fsm_unused_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	struct gsm_bts *bts = lchan->ts->trx->bts;
 	lchan_reset(lchan);
 	osmo_fsm_inst_dispatch(lchan->ts->fi, TS_EV_LCHAN_UNUSED, lchan);
+
+	/* Poll the channel request queue, so that waiting calls can make use of the lchan that just
+	 * has become unused now. */
+	abis_rsl_chan_rqd_queue_poll(bts);
+}
+
+static void lchan_fsm_wait_after_error_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	struct gsm_bts *bts = lchan->ts->trx->bts;
+
+	/* We also need to poll the channel request queue when the FSM enters the WAIT_AFTER_ERROR
+	 * state. In case of an emergency call the channel request queue will skip the waiting
+	 * period. */
+	abis_rsl_chan_rqd_queue_poll(bts);
 }
 
 /* Configure the multirate setting on this channel. */
-static int lchan_mr_config(struct gsm_lchan *lchan, const struct gsm48_multi_rate_conf *mr_conf)
+static int lchan_mr_config(struct gsm_lchan *lchan, uint16_t s15_s0)
 {
+	struct gsm48_multi_rate_conf mr_conf;
 	bool full_rate = (lchan->type == GSM_LCHAN_TCH_F);
 	struct gsm_bts *bts = lchan->ts->trx->bts;
-	struct bsc_msc_data *msc = lchan->conn->sccp.msc;
 	struct amr_multirate_conf *mr;
 	int rc;
 	int rc_rate;
 	struct gsm48_multi_rate_conf mr_conf_filtered;
 	const struct gsm48_multi_rate_conf *mr_conf_bts;
 
+	/* Generate mr conf struct from S15-S0 bits */
+	if (gsm48_mr_cfg_from_gsm0808_sc_cfg(&mr_conf, s15_s0) < 0) {
+		LOG_LCHAN(lchan, LOGL_ERROR,
+			  "can not determine multirate configuration, S15-S0 (%04x) are ambiguous!\n", s15_s0);
+		return -EINVAL;
+	}
+
+	/* Do not include 12.2 kbps rate when S1 is set. */
+	if (lchan->type == GSM_LCHAN_TCH_H && (s15_s0 & GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20)) {
+		/* See also 3GPP TS 28.062, chapter 7.11.3.1.3: "In case this Configuration
+		 * "Config-NB-Code = 1" is signalled in the TFO Negotiation for the HR_AMR
+		 * Codec Type,then it shall be assumed that AMR mode 12.2 kbps is (of course)
+		 * not included. */
+		mr_conf.m12_2 = 0;
+	}
+
 	/* There are two different active sets, depending on the channel rate,
 	 * make sure the appropate one is selected. */
 	if (full_rate)
 		mr = &bts->mr_full;
 	else
 		mr = &bts->mr_half;
+	mr_conf_bts = (struct gsm48_multi_rate_conf *)mr->gsm48_ie;
 
-	/* The VTY allows to forbid certain codec rates. Unfortunately we can
-	 * not articulate all of the prohibitions on through S0-S15 on the A
-	 * interface. To ensure that the VTY settings are observed we create
-	 * a manipulated copy of the mr_conf that ensures forbidden codec rates
-	 * are not used in the multirate configuration IE. */
-	rc_rate = calc_amr_rate_intersection(&mr_conf_filtered, &msc->amr_conf, mr_conf);
-	if (rc_rate < 0) {
-		LOG_LCHAN(lchan, LOGL_ERROR,
-			  "can not encode multirate configuration (invalid amr rate setting, MSC)\n");
-		return -EINVAL;
+	if (lchan->activate.info.activ_for == FOR_VTY)
+		/* If the channel is activated manually from VTY, then there is no
+		 * conn attached to the lchan, also no MSC is involved. Since this
+		 * option is for debugging and the codec choice is an intentional
+		 * decision by the VTY user, we do not filter the mr_conf. */
+		memcpy(&mr_conf_filtered, &mr_conf, sizeof(mr_conf_filtered));
+	else {
+		/* The VTY allows to forbid certain codec rates. Unfortunately we can
+		 * not articulate all of the prohibitions on through S0-S15 on the A
+		 * interface. To ensure that the VTY settings are observed we create
+		 * a manipulated copy of the mr_conf that ensures forbidden codec rates
+		 * are not used in the multirate configuration IE. */
+		rc_rate = calc_amr_rate_intersection(&mr_conf_filtered, &lchan->conn->sccp.msc->amr_conf, &mr_conf);
+		if (rc_rate < 0) {
+			LOG_LCHAN(lchan, LOGL_ERROR,
+				  "can not encode multirate configuration (invalid amr rate setting, MSC)\n");
+			return -EINVAL;
+		}
 	}
 
 	/* The two last codec rates which are defined for AMR do only work with
@@ -454,15 +500,23 @@ static int lchan_mr_config(struct gsm_lchan *lchan, const struct gsm48_multi_rat
 	}
 
 	/* Ensure that the resulting filtered conf is coherent with the
-	 * configuration that is set for the BTS and the specified rate */
-	mr_conf_bts = (struct gsm48_multi_rate_conf *)mr->gsm48_ie;
-	rc_rate = calc_amr_rate_intersection(&mr_conf_filtered, mr_conf_bts, &mr_conf_filtered);
-	if (rc_rate < 0) {
-		LOG_LCHAN(lchan, LOGL_ERROR,
-			  "can not encode multirate configuration (invalid amr rate setting, BTS)\n");
-		return -EINVAL;
+	 * configuration that is set for the BTS and the specified rate.
+	 * if the channel activation was triggerd by the VTY, do not
+	 * filter anything (see also comment above) */
+	if (lchan->activate.info.activ_for != FOR_VTY) {
+		rc_rate = calc_amr_rate_intersection(&mr_conf_filtered, mr_conf_bts, &mr_conf_filtered);
+		if (rc_rate < 0) {
+			LOG_LCHAN(lchan, LOGL_ERROR,
+				  "can not encode multirate configuration (invalid amr rate setting, BTS)\n");
+			return -EINVAL;
+		}
 	}
 
+	/* Set the ICMI according to the BTS. Above gsm48_mr_cfg_from_gsm0808_sc_cfg() always sets ICMI = 1, which
+	 * carried through all of the above rate intersections. */
+	mr_conf_filtered.icmi = mr_conf_bts->icmi;
+	mr_conf_filtered.smod = mr_conf_bts->smod;
+
 	/* Proceed with the generation of the multirate configuration IE
 	 * (MS and BTS) */
 	rc = gsm48_multirate_config(lchan->mr_ms_lv, &mr_conf_filtered, mr->ms_mode, mr->num_modes);
@@ -476,6 +530,7 @@ static int lchan_mr_config(struct gsm_lchan *lchan, const struct gsm48_multi_rat
 		return -EINVAL;
 	}
 
+	lchan->s15_s0 = s15_s0;
 	return 0;
 }
 
@@ -506,7 +561,6 @@ static void lchan_fsm_unused(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 static void lchan_fsm_wait_ts_ready_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
-	struct gsm48_multi_rate_conf mr_conf;
 	struct gsm_bts *bts = lchan->ts->trx->bts;
 	struct osmo_mgcpc_ep_ci *use_mgwep_ci;
 	struct gsm_lchan *old_lchan = lchan->activate.info.re_use_mgw_endpoint_from_lchan;
@@ -526,35 +580,24 @@ static void lchan_fsm_wait_ts_ready_onenter(struct osmo_fsm_inst *fi, uint32_t p
 		ms_power_dbm = ms_pwr_dbm(bts->band, old_lchan->ms_power);
 		lchan_update_ms_power_ctrl_level(lchan, ms_power_dbm >= 0 ? ms_power_dbm : bts->ms_max_power);
 		lchan->bs_power = old_lchan->bs_power;
-		lchan->rqd_ta = old_lchan->rqd_ta;
 	} else {
 		lchan_update_ms_power_ctrl_level(lchan, bts->ms_max_power);
 		/* Upon last entering the UNUSED state, from lchan_reset():
 		 * - bs_power is still zero, 0dB reduction, output power = Pn.
 		 * - TA is still zero, to be determined by RACH. */
+
+		/* Default BS Power reduction value (in 2 dB steps) */
+		if (bts->bs_power_ctrl.mode == GSM_PWR_CTRL_MODE_DYN_BTS)
+			lchan->bs_power = bts->bs_power_ctrl.bs_power_max_db / 2;
+		else
+			lchan->bs_power = bts->bs_power_ctrl.bs_power_val_db / 2;
 	}
 
 	if (info->chan_mode == GSM48_CMODE_SPEECH_AMR) {
-		if (gsm48_mr_cfg_from_gsm0808_sc_cfg(&mr_conf, info->s15_s0) < 0) {
-			lchan_fail("Can not determine multirate configuration, S15-S0 (%04x) are ambiguous!\n",
-				   info->s15_s0);
-			return;
-		}
-
-		/* Do not include 12.2 kbps rate when S1 is set. */
-		if (lchan->type == GSM_LCHAN_TCH_H && (info->s15_s0 & GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20)) {
-			/* See also 3GPP TS 28.062, chapter 7.11.3.1.3: "In case this Configuration
-			 * "Config-NB-Code = 1" is signalled in the TFO Negotiation for the HR_AMR
-			 * Codec Type,then it shall be assumed that AMR mode 12.2 kbps is (of course)
-			 * not included. */
-			mr_conf.m12_2 = 0;
-		}
-
-		if (lchan_mr_config(lchan, &mr_conf) < 0) {
+		if (lchan_mr_config(lchan, info->s15_s0) < 0) {
 			lchan_fail("Can not generate multirate configuration IE\n");
 			return;
 		}
-		lchan->s15_s0 = info->s15_s0;
 	}
 
 	switch (info->chan_mode) {
@@ -658,8 +701,13 @@ static void lchan_fsm_wait_activ_ack_onenter(struct osmo_fsm_inst *fi, uint32_t
 	lchan->encr = lchan->activate.info.encr;
 
 	rc = rsl_tx_chan_activ(lchan, act_type, ho_ref);
-	if (rc)
+	if (rc) {
 		lchan_fail_to(LCHAN_ST_UNUSED, "Tx Chan Activ failed: %s (%d)", strerror(-rc), rc);
+		return;
+	}
+
+	if (lchan->activate.info.ta_known)
+		lchan->last_ta = lchan->activate.info.ta;
 }
 
 static void lchan_fsm_post_activ_ack(struct osmo_fsm_inst *fi);
@@ -679,6 +727,7 @@ static void lchan_fsm_wait_activ_ack(struct osmo_fsm_inst *fi, uint32_t event, v
 		if (data) {
 			uint32_t next_state;
 			lchan->release.rsl_error_cause = *(uint8_t*)data;
+			lchan->release.rr_cause = bsc_gsm48_rr_cause_from_rsl_cause(lchan->release.rsl_error_cause);
 			lchan->release.in_error = true;
 			if (lchan->release.rsl_error_cause != RSL_ERR_RCH_ALR_ACTV_ALLOC)
 				next_state = LCHAN_ST_BORKEN;
@@ -691,6 +740,7 @@ static void lchan_fsm_wait_activ_ack(struct osmo_fsm_inst *fi, uint32_t event, v
 				      rsl_err_name(lchan->release.rsl_error_cause), lchan->release.rsl_error_cause);
 		} else {
 			lchan->release.rsl_error_cause = RSL_ERR_IE_NONEXIST;
+			lchan->release.rr_cause = bsc_gsm48_rr_cause_from_rsl_cause(lchan->release.rsl_error_cause);
 			lchan->release.in_error = true;
 			lchan_fail_to(LCHAN_ST_BORKEN, "Chan Activ NACK without cause IE");
 		}
@@ -792,6 +842,16 @@ static void lchan_fsm_wait_rll_rtp_establish_onenter(struct osmo_fsm_inst *fi, u
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
 	if (lchan->fi_rtp)
 		osmo_fsm_inst_dispatch(lchan->fi_rtp, LCHAN_RTP_EV_LCHAN_READY, 0);
+	/* Prepare an MGW endpoint CI if appropriate (late). */
+	else if (lchan->activate.info.requires_voice_stream)
+		lchan_rtp_fsm_start(lchan);
+
+	/* When activating a channel for VTY, skip waiting for activity from
+	 * lchan_rtp_fsm, but only if no voice stream is required. */
+	if (lchan->activate.info.activ_for == FOR_VTY &&
+	    !lchan->activate.info.requires_voice_stream) {
+		lchan_fsm_state_chg(LCHAN_ST_ESTABLISHED);
+	}
 }
 
 static void lchan_fsm_wait_rll_rtp_establish(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -828,6 +888,66 @@ static void lchan_fsm_wait_rll_rtp_establish(struct osmo_fsm_inst *fi, uint32_t
 	}
 }
 
+static void lchan_fsm_wait_rr_chan_mode_modify_ack_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	gsm48_lchan_modify(lchan, lchan->activate.info.chan_mode);
+}
+
+static void lchan_fsm_wait_rr_chan_mode_modify_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+{
+	switch (event) {
+
+	case LCHAN_EV_RR_CHAN_MODE_MODIFY_ACK:
+		lchan_fsm_state_chg(LCHAN_ST_WAIT_RSL_CHAN_MODE_MODIFY_ACK);
+		return;
+
+	case LCHAN_EV_RR_CHAN_MODE_MODIFY_ERROR:
+		lchan_fail("Failed to change channel mode on the MS side: %s in state %s\n",
+			   osmo_fsm_event_name(fi->fsm, event),
+			   osmo_fsm_inst_state_name(fi));
+		return;
+
+	default:
+		OSMO_ASSERT(false);
+	}
+}
+
+static void lchan_fsm_wait_rsl_chan_mode_modify_ack_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	int rc;
+
+	rc = rsl_chan_mode_modify_req(lchan);
+	if (rc < 0) {
+		lchan_fail("Failed to send rsl message to change the channel mode on the BTS side: state %s\n",
+			   osmo_fsm_inst_state_name(fi));
+	}
+}
+
+static void lchan_fsm_wait_rsl_chan_mode_modify_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+{
+	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	switch (event) {
+
+	case LCHAN_EV_RSL_CHAN_MODE_MODIFY_ACK:
+		if (lchan->activate.info.requires_voice_stream)
+			lchan_fsm_state_chg(LCHAN_ST_WAIT_RLL_RTP_ESTABLISH);
+		else
+			lchan_fsm_state_chg(LCHAN_ST_ESTABLISHED);
+		return;
+
+	case LCHAN_EV_RSL_CHAN_MODE_MODIFY_NACK:
+		lchan_fail("Failed to change channel mode on the BTS side: %s in state %s\n",
+			   osmo_fsm_event_name(fi->fsm, event),
+			   osmo_fsm_inst_state_name(fi));
+		return;
+
+	default:
+		OSMO_ASSERT(false);
+	}
+}
+
 static void lchan_fsm_established_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
@@ -904,6 +1024,8 @@ static void handle_rll_rel_ind_or_conf(struct osmo_fsm_inst *fi, uint32_t event,
 static void lchan_fsm_established(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	struct lchan_activate_info *info;
+	struct osmo_mgcpc_ep_ci *use_mgwep_ci;
 
 	switch (event) {
 	case LCHAN_EV_RLL_ESTABLISH_IND:
@@ -930,6 +1052,48 @@ static void lchan_fsm_established(struct osmo_fsm_inst *fi, uint32_t event, void
 			   osmo_fsm_inst_state_name(fi));
 		return;
 
+	case LCHAN_EV_REQUEST_MODE_MODIFY:
+
+		/* FIXME: Add missing implementation to handle an already existing RTP voice stream on MODE MODIFY.
+		 * there may be transitions from VOICE to SIGNALLING and also from VOICE to VOICE with a different
+		 * codec. */
+		if (lchan->fi_rtp) {
+			lchan_fail("MODE MODIFY not implemented when RTP voice stream is already active (VOICE => SIGNALLING, VOICE/CODEC_A => VOICE/CODEC_B)\n");
+			return;
+		}
+
+		info = data;
+		lchan->activate.info = *info;
+		use_mgwep_ci = lchan_use_mgw_endpoint_ci_bts(lchan);
+
+		if (info->chan_mode == GSM48_CMODE_SPEECH_AMR) {
+			if (lchan_mr_config(lchan, info->s15_s0) < 0) {
+				lchan_fail("Can not generate multirate configuration IE\n");
+				return;
+			}
+		}
+
+		LOG_LCHAN(lchan, LOGL_INFO,
+			  "Modification requested: %s voice=%s MGW-ci=%s type=%s tch-mode=%s encr-alg=A5/%u ck=%s\n",
+			  lchan_activate_mode_name(lchan->activate.info.activ_for),
+			  lchan->activate.info.requires_voice_stream ? "yes" : "no",
+			  lchan->activate.info.requires_voice_stream ?
+			  (use_mgwep_ci ? osmo_mgcpc_ep_ci_name(use_mgwep_ci) : "new")
+			  : "none",
+			  gsm_lchant_name(lchan->type),
+			  gsm48_chan_mode_name(lchan->tch_mode),
+			  (lchan->activate.info.encr.alg_id ? : 1) - 1,
+			  lchan->activate.info.encr.key_len ? osmo_hexdump_nospc(lchan->activate.info.encr.key,
+										 lchan->activate.info.encr.key_len) : "none");
+
+		/* While the mode is changed the lchan is virtually "not activated", at least
+		 * from the FSM implementations perspective */
+		lchan->activate.concluded = false;
+
+		/* Initiate mode modification, start with the MS side (RR) */
+		lchan_fsm_state_chg(LCHAN_ST_WAIT_RR_CHAN_MODE_MODIFY_ACK);
+		return;
+
 	default:
 		OSMO_ASSERT(false);
 	}
@@ -1015,7 +1179,7 @@ static void lchan_fsm_wait_rll_rtp_released(struct osmo_fsm_inst *fi, uint32_t e
 		 * TODO: that's how the code was before lchan FSM, is this correct/useful? */
 		handle_rll_rel_ind_or_conf(fi, event, data);
 		break;
-	
+
 	case LCHAN_EV_RTP_RELEASED:
 	case LCHAN_EV_RTP_ERROR:
 		break;
@@ -1071,30 +1235,68 @@ static void lchan_fsm_wait_rf_release_ack(struct osmo_fsm_inst *fi, uint32_t eve
 static void lchan_fsm_borken_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	struct gsm_bts *bts = lchan->ts->trx->bts;
+	enum bts_counter_id ctr;
+	switch (prev_state) {
+	case LCHAN_ST_UNUSED:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_UNUSED;
+		break;
+	case LCHAN_ST_WAIT_ACTIV_ACK:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_WAIT_ACTIV_ACK;
+		break;
+	case LCHAN_ST_WAIT_RF_RELEASE_ACK:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_WAIT_RF_RELEASE_ACK;
+		break;
+	case LCHAN_ST_BORKEN:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_BORKEN;
+		break;
+	case LCHAN_ST_WAIT_RR_CHAN_MODE_MODIFY_ACK:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_WAIT_RR_CHAN_MODE_MODIFY_ACK;
+		break;
+	case LCHAN_ST_WAIT_RSL_CHAN_MODE_MODIFY_ACK:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_WAIT_RSL_CHAN_MODE_MODIFY_ACK;
+		break;
+	default:
+		ctr = BTS_CTR_LCHAN_BORKEN_FROM_UNKNOWN;
+	}
+	rate_ctr_inc(&bts->bts_ctrs->ctr[ctr]);
+	if (prev_state != LCHAN_ST_BORKEN)
+		osmo_stat_item_inc(bts->bts_statg->items[BTS_STAT_LCHAN_BORKEN], 1);
+
+	/* The actual action besides all the beancounting above */
 	lchan_reset(lchan);
 }
 
 static void lchan_fsm_borken(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	struct gsm_bts *bts = lchan->ts->trx->bts;
 	switch (event) {
 
 	case LCHAN_EV_RSL_CHAN_ACTIV_ACK:
 		/* A late Chan Activ ACK? Release. */
+		rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_LCHAN_BORKEN_EV_CHAN_ACTIV_ACK]);
+		osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_LCHAN_BORKEN], 1);
 		lchan->release.in_error = true;
 		lchan->release.rsl_error_cause = RSL_ERR_INTERWORKING;
+		lchan->release.rr_cause = bsc_gsm48_rr_cause_from_rsl_cause(lchan->release.rsl_error_cause);
 		lchan_fsm_state_chg(LCHAN_ST_WAIT_RF_RELEASE_ACK);
 		return;
 
 	case LCHAN_EV_RSL_CHAN_ACTIV_NACK:
 		/* A late Chan Activ NACK? Ok then, unused. */
+		rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_LCHAN_BORKEN_EV_CHAN_ACTIV_NACK]);
+		osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_LCHAN_BORKEN], 1);
 		lchan_fsm_state_chg(LCHAN_ST_UNUSED);
 		return;
 
 	case LCHAN_EV_RSL_RF_CHAN_REL_ACK:
 		/* A late Release ACK? */
+		rate_ctr_inc(&bts->bts_ctrs->ctr[BTS_CTR_LCHAN_BORKEN_EV_RF_CHAN_REL_ACK]);
+		osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_LCHAN_BORKEN], 1);
 		lchan->release.in_error = true;
 		lchan->release.rsl_error_cause = RSL_ERR_INTERWORKING;
+		lchan->release.rr_cause = bsc_gsm48_rr_cause_from_rsl_cause(lchan->release.rsl_error_cause);
 		lchan_fsm_state_chg(LCHAN_ST_WAIT_AFTER_ERROR);
 		/* TODO: we used to do this only for sysmobts:
 			int do_free = is_sysmobts_v2(ts->trx->bts);
@@ -1188,6 +1390,32 @@ static const struct osmo_fsm_state lchan_fsm_states[] = {
 			| S(LCHAN_ST_WAIT_RLL_RTP_RELEASED)
 			,
 	},
+	[LCHAN_ST_WAIT_RR_CHAN_MODE_MODIFY_ACK] = {
+		.name = "WAIT_CHAN_RR_MODE_MODIFY_ACK",
+		.onenter = lchan_fsm_wait_rr_chan_mode_modify_ack_onenter,
+		.action = lchan_fsm_wait_rr_chan_mode_modify_ack,
+		.in_event_mask = 0
+			| S(LCHAN_EV_RR_CHAN_MODE_MODIFY_ACK)
+			| S(LCHAN_EV_RR_CHAN_MODE_MODIFY_ERROR)
+			,
+		.out_state_mask = 0
+			| S(LCHAN_ST_BORKEN)
+			| S(LCHAN_ST_WAIT_RSL_CHAN_MODE_MODIFY_ACK)
+			,
+	},
+	[LCHAN_ST_WAIT_RSL_CHAN_MODE_MODIFY_ACK] = {
+		.name = "WAIT_RSL_CHAN_MODE_MODIFY_ACK",
+		.onenter = lchan_fsm_wait_rsl_chan_mode_modify_ack_onenter,
+		.action = lchan_fsm_wait_rsl_chan_mode_modify_ack,
+		.in_event_mask = 0
+			| S(LCHAN_EV_RSL_CHAN_MODE_MODIFY_ACK)
+			| S(LCHAN_EV_RSL_CHAN_MODE_MODIFY_NACK)
+			,
+		.out_state_mask = 0
+			| S(LCHAN_ST_BORKEN)
+			| S(LCHAN_ST_WAIT_RLL_RTP_ESTABLISH)
+			,
+	},
 	[LCHAN_ST_ESTABLISHED] = {
 		.name = "ESTABLISHED",
 		.onenter = lchan_fsm_established_onenter,
@@ -1198,12 +1426,14 @@ static const struct osmo_fsm_state lchan_fsm_states[] = {
 			| S(LCHAN_EV_RLL_ESTABLISH_IND) /* ignored */
 			| S(LCHAN_EV_RTP_ERROR)
 			| S(LCHAN_EV_RTP_RELEASED)
+			| S(LCHAN_EV_REQUEST_MODE_MODIFY)
 			,
 		.out_state_mask = 0
 			| S(LCHAN_ST_UNUSED)
 			| S(LCHAN_ST_WAIT_RLL_RTP_RELEASED)
 			| S(LCHAN_ST_WAIT_BEFORE_RF_RELEASE)
 			| S(LCHAN_ST_WAIT_RF_RELEASE_ACK)
+			| S(LCHAN_ST_WAIT_RR_CHAN_MODE_MODIFY_ACK)
 			,
 	},
 	[LCHAN_ST_WAIT_RLL_RTP_RELEASED] = {
@@ -1249,6 +1479,7 @@ static const struct osmo_fsm_state lchan_fsm_states[] = {
 	},
 	[LCHAN_ST_WAIT_AFTER_ERROR] = {
 		.name = "WAIT_AFTER_ERROR",
+		.onenter = lchan_fsm_wait_after_error_onenter,
 		.in_event_mask = 0
 			| S(LCHAN_EV_RTP_RELEASED) /* ignore late lchan_rtp_fsm release events */
 			,
@@ -1268,6 +1499,7 @@ static const struct osmo_fsm_state lchan_fsm_states[] = {
 			| S(LCHAN_EV_RTP_RELEASED)
 			,
 		.out_state_mask = 0
+			| S(LCHAN_ST_WAIT_RF_RELEASE_ACK)
 			| S(LCHAN_ST_UNUSED)
 			| S(LCHAN_ST_WAIT_AFTER_ERROR)
 			,
@@ -1288,12 +1520,14 @@ static const struct value_string lchan_fsm_event_names[] = {
 	OSMO_VALUE_STRING(LCHAN_EV_RLL_REL_CONF),
 	OSMO_VALUE_STRING(LCHAN_EV_RSL_RF_CHAN_REL_ACK),
 	OSMO_VALUE_STRING(LCHAN_EV_RLL_ERR_IND),
-	OSMO_VALUE_STRING(LCHAN_EV_CHAN_MODE_MODIF_ACK),
-	OSMO_VALUE_STRING(LCHAN_EV_CHAN_MODE_MODIF_ERROR),
+	OSMO_VALUE_STRING(LCHAN_EV_RR_CHAN_MODE_MODIFY_ACK),
+	OSMO_VALUE_STRING(LCHAN_EV_RR_CHAN_MODE_MODIFY_ERROR),
+	OSMO_VALUE_STRING(LCHAN_EV_RSL_CHAN_MODE_MODIFY_ACK),
+	OSMO_VALUE_STRING(LCHAN_EV_RSL_CHAN_MODE_MODIFY_NACK),
 	{}
 };
 
-void lchan_fsm_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void lchan_fsm_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	switch (event) {
 
@@ -1301,12 +1535,26 @@ void lchan_fsm_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *d
 		lchan_fail_to(LCHAN_ST_UNUSED, "LCHAN_EV_TS_ERROR");
 		return;
 
+	case LCHAN_EV_RLL_ERR_IND:
+		/* let's just ignore this.  We are already logging the
+		 * fact that this message was received inside
+		 * abis_rsl.c.  There can be any number of reasons why the
+		 * radio link layer failed */
+		return;
+
 	default:
 		return;
 	}
 }
 
-int lchan_fsm_timer_cb(struct osmo_fsm_inst *fi)
+void lchan_fsm_skip_error(struct gsm_lchan *lchan)
+{
+	struct osmo_fsm_inst *fi = lchan->fi;
+	if (fi->state == LCHAN_ST_WAIT_AFTER_ERROR)
+		lchan_fsm_state_chg(LCHAN_ST_UNUSED);
+}
+
+static int lchan_fsm_timer_cb(struct osmo_fsm_inst *fi)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
 	switch (fi->state) {
@@ -1322,6 +1570,7 @@ int lchan_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	default:
 		lchan->release.in_error = true;
 		lchan->release.rsl_error_cause = RSL_ERR_INTERWORKING;
+		lchan->release.rr_cause = bsc_gsm48_rr_cause_from_rsl_cause(lchan->release.rsl_error_cause);
 		lchan_fail("Timeout");
 		return 0;
 	}
@@ -1340,8 +1589,8 @@ void lchan_release(struct gsm_lchan *lchan, bool do_rr_release,
 	struct osmo_fsm_inst *fi = lchan->fi;
 
 	lchan->release.in_error = err;
-	lchan->release.rsl_error_cause = cause_rr;
 	lchan->release.do_rr_release = do_rr_release;
+	lchan->release.rr_cause = cause_rr;
 
 	/* States waiting for events will notice the desire to release when done waiting, so it is enough
 	 * to mark for release. */
@@ -1380,9 +1629,13 @@ exit_release_handler:
 	lchan->release.in_release_handler = false;
 }
 
-void lchan_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
+static void lchan_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	if (lchan->fi->state == LCHAN_ST_BORKEN) {
+		rate_ctr_inc(&lchan->ts->trx->bts->bts_ctrs->ctr[BTS_CTR_LCHAN_BORKEN_EV_TEARDOWN]);
+		osmo_stat_item_dec(lchan->ts->trx->bts->bts_statg->items[BTS_STAT_LCHAN_BORKEN], 1);
+	}
 	lchan_reset(lchan);
 	if (lchan->last_error) {
 		talloc_free(lchan->last_error);
@@ -1424,6 +1677,7 @@ static struct osmo_fsm lchan_fsm = {
 	.allstate_action = lchan_fsm_allstate_action,
 	.allstate_event_mask = 0
 		| S(LCHAN_EV_TS_ERROR)
+		| S(LCHAN_EV_RLL_ERR_IND)
 		,
 	.timer_cb = lchan_fsm_timer_cb,
 	.cleanup = lchan_fsm_cleanup,
