@@ -30,6 +30,7 @@
 #include <osmocom/bsc/bsc_subscr_conn_fsm.h>
 #include <osmocom/bsc/abis_rsl.h>
 #include <osmocom/bsc/bsc_msc_data.h>
+#include <osmocom/bsc/bts.h>
 
 static struct osmo_fsm lchan_rtp_fsm;
 
@@ -42,10 +43,10 @@ struct gsm_lchan *lchan_rtp_fi_lchan(struct osmo_fsm_inst *fi)
 }
 
 struct osmo_tdef_state_timeout lchan_rtp_fsm_timeouts[32] = {
-	[LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_AVAILABLE] = { .T=23004 },
-	[LCHAN_RTP_ST_WAIT_IPACC_CRCX_ACK]	= { .T=23005 },
-	[LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK]	= { .T=23006 },
-	[LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_CONFIGURED] = { .T=23004 },
+	[LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_AVAILABLE] = { .T=-9 },
+	[LCHAN_RTP_ST_WAIT_IPACC_CRCX_ACK]	= { .T=-7 },
+	[LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK]	= { .T=-8 },
+	[LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_CONFIGURED] = { .T=-10 },
 };
 
 /* Transition to a state, using the T timer defined in lchan_rtp_fsm_timeouts.
@@ -140,6 +141,12 @@ static void lchan_rtp_fsm_wait_mgw_endpoint_available_onenter(struct osmo_fsm_in
 	struct osmo_mgcpc_ep_ci *use_mgwep_ci = lchan_use_mgw_endpoint_ci_bts(lchan);
 	struct mgcp_conn_peer crcx_info = {};
 
+	if (!is_ipaccess_bts(lchan->ts->trx->bts)) {
+		LOG_LCHAN_RTP(lchan, LOGL_DEBUG, "Audio link to-BTS via E1, skipping IPACC\n");
+		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_WAIT_LCHAN_READY);
+		return;
+	}
+
 	if (use_mgwep_ci) {
 		LOG_LCHAN_RTP(lchan, LOGL_DEBUG, "MGW endpoint already available: %s\n",
 			      osmo_mgcpc_ep_ci_name(use_mgwep_ci));
@@ -147,7 +154,7 @@ static void lchan_rtp_fsm_wait_mgw_endpoint_available_onenter(struct osmo_fsm_in
 		return;
 	}
 
-	mgwep = gscon_ensure_mgw_endpoint(lchan->conn, lchan->activate.info.msc_assigned_cic);
+	mgwep = gscon_ensure_mgw_endpoint(lchan->conn, lchan->activate.info.msc_assigned_cic, lchan);
 	if (!mgwep) {
 		lchan_rtp_fail("Internal error: cannot obtain MGW endpoint handle for conn");
 		return;
@@ -247,7 +254,7 @@ static void lchan_rtp_fsm_post_lchan_ready(struct osmo_fsm_inst *fi)
 	if (is_ipaccess_bts(lchan->ts->trx->bts))
 		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_WAIT_IPACC_CRCX_ACK);
 	else
-		lchan_rtp_fsm_switch_rtp(fi);
+		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_CONFIGURED);
 }
 
 static void lchan_rtp_fsm_wait_ipacc_crcx_ack_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
@@ -294,10 +301,7 @@ static void lchan_rtp_fsm_wait_ipacc_crcx_ack(struct osmo_fsm_inst *fi, uint32_t
 	switch (event) {
 
 	case LCHAN_RTP_EV_IPACC_CRCX_ACK:
-		/* the CRCX ACK parsing has already noted the RTP port information at
-		 * lchan->abis_ip.bound_*, see ipac_parse_rtp(). We'll use that in
-		 * lchan_rtp_fsm_wait_mgw_endpoint_configured_onenter(). */
-		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK);
+		lchan_rtp_fsm_switch_rtp(fi);
 		return;
 
 	case LCHAN_RTP_EV_IPACC_CRCX_NACK:
@@ -323,6 +327,7 @@ static void lchan_rtp_fsm_wait_ipacc_mdcx_ack_onenter(struct osmo_fsm_inst *fi,
 	int rc;
 	struct gsm_lchan *lchan = lchan_rtp_fi_lchan(fi);
 	const struct mgcp_conn_peer *mgw_rtp;
+	struct in_addr sin;
 
 	if (lchan->release.requested) {
 		lchan_rtp_fail("Release requested while activating");
@@ -337,8 +342,13 @@ static void lchan_rtp_fsm_wait_ipacc_mdcx_ack_onenter(struct osmo_fsm_inst *fi,
 		return;
 	}
 
-	/* Other RTP settings were already setup in lchan_rtp_fsm_wait_ipacc_crcx_ack_onenter() */
-	lchan->abis_ip.connect_ip = ntohl(inet_addr(mgw_rtp->addr));
+	/* Other RTP settings were already set up in lchan_rtp_fsm_wait_ipacc_crcx_ack_onenter() */
+	if (inet_pton(AF_INET, mgw_rtp->addr, &sin) != 1) {
+		/* Only IPv4 addresses are supported in IPACC */
+		lchan_rtp_fail("Invalid remote IPv4 address %s", mgw_rtp->addr);
+		return;
+	}
+	lchan->abis_ip.connect_ip = ntohl(sin.s_addr);
 	lchan->abis_ip.connect_port = mgw_rtp->port;
 
 	/* send-recv */
@@ -353,21 +363,16 @@ static void lchan_rtp_fsm_wait_ipacc_mdcx_ack_onenter(struct osmo_fsm_inst *fi,
 
 static void lchan_rtp_fsm_wait_ipacc_mdcx_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
-	struct gsm_lchan *lchan = lchan_rtp_fi_lchan(fi);
 	switch (event) {
 
 	case LCHAN_RTP_EV_IPACC_MDCX_ACK:
-		lchan_rtp_fsm_switch_rtp(fi);
+		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_READY);
 		return;
 
 	case LCHAN_RTP_EV_IPACC_MDCX_NACK:
 		lchan_rtp_fail("Received NACK on IPACC MDCX");
 		return;
 
-	case LCHAN_RTP_EV_READY_TO_SWITCH_RTP:
-		lchan->activate.info.wait_before_switching_rtp = false;
-		return;
-
 	case LCHAN_RTP_EV_RELEASE:
 	case LCHAN_RTP_EV_ROLLBACK:
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REQUEST, 0);
@@ -444,6 +449,12 @@ static void lchan_rtp_fsm_wait_mgw_endpoint_configured_onenter(struct osmo_fsm_i
 		return;
 	}
 
+	if (!is_ipaccess_bts(lchan->ts->trx->bts)) {
+		LOG_LCHAN_RTP(lchan, LOGL_DEBUG, "Audio link to-BTS via E1, skipping IPACC\n");
+		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_READY);
+		return;
+	}
+
 	/* At this point, we are taking over an old lchan's MGW endpoint (if any). */
 	if (!lchan->mgw_endpoint_ci_bts && old_lchan) {
 		/* The old lchan shall forget the endpoint now. We might put it back upon ROLLBACK */
@@ -461,10 +472,15 @@ static void lchan_rtp_fsm_wait_mgw_endpoint_configured_onenter(struct osmo_fsm_i
 
 static void lchan_rtp_fsm_wait_mgw_endpoint_configured(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
-	switch (event) {
+	struct gsm_lchan *lchan = lchan_rtp_fi_lchan(fi);
 
+	switch (event) {
 	case LCHAN_RTP_EV_MGW_ENDPOINT_CONFIGURED:
-		lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_READY);
+		if (is_ipaccess_bts(lchan->ts->trx->bts))
+			lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK);
+		else {
+			lchan_rtp_fsm_state_chg(LCHAN_RTP_ST_READY);
+		}
 		return;
 
 	case LCHAN_RTP_EV_MGW_ENDPOINT_ERROR:
@@ -520,7 +536,11 @@ static void lchan_rtp_fsm_rollback_onenter(struct osmo_fsm_inst *fi, uint32_t pr
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REQUEST, 0);
 		return;
 	}
-	connect_mgw_endpoint_to_lchan(fi, lchan->mgw_endpoint_ci_bts, old_lchan);
+
+	if (is_ipaccess_bts(lchan->ts->trx->bts))
+		connect_mgw_endpoint_to_lchan(fi, lchan->mgw_endpoint_ci_bts, old_lchan);
+	else
+		osmo_fsm_inst_dispatch(fi, LCHAN_RTP_EV_MGW_ENDPOINT_CONFIGURED, 0);
 }
 
 static void lchan_rtp_fsm_rollback(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -628,24 +648,9 @@ static const struct osmo_fsm_state lchan_rtp_fsm_states[] = {
 			| S(LCHAN_RTP_EV_RELEASE)
 			| S(LCHAN_RTP_EV_ROLLBACK)
 			,
-		.out_state_mask = 0
-			| S(LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK)
-			,
-	},
-	[LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK] = {
-		.name = "WAIT_IPACC_MDCX_ACK",
-		.onenter = lchan_rtp_fsm_wait_ipacc_mdcx_ack_onenter,
-		.action = lchan_rtp_fsm_wait_ipacc_mdcx_ack,
-		.in_event_mask = 0
-			| S(LCHAN_RTP_EV_READY_TO_SWITCH_RTP)
-			| S(LCHAN_RTP_EV_IPACC_MDCX_ACK)
-			| S(LCHAN_RTP_EV_IPACC_MDCX_NACK)
-			| S(LCHAN_RTP_EV_RELEASE)
-			| S(LCHAN_RTP_EV_ROLLBACK)
-			,
 		.out_state_mask = 0
 			| S(LCHAN_RTP_ST_WAIT_READY_TO_SWITCH_RTP)
-			| S(LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_CONFIGURED)
+			| S(LCHAN_RTP_ST_WAIT_MGW_ENDPOINT_CONFIGURED) /*old: LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK*/
 			,
 	},
 	[LCHAN_RTP_ST_WAIT_READY_TO_SWITCH_RTP] = {
@@ -671,7 +676,24 @@ static const struct osmo_fsm_state lchan_rtp_fsm_states[] = {
 			| S(LCHAN_RTP_EV_ROLLBACK)
 			,
 		.out_state_mask = 0
+			| S(LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK)
 			| S(LCHAN_RTP_ST_READY)
+			| S(LCHAN_RTP_ST_ROLLBACK)
+			,
+	},
+	[LCHAN_RTP_ST_WAIT_IPACC_MDCX_ACK] = {
+		.name = "WAIT_IPACC_MDCX_ACK",
+		.onenter = lchan_rtp_fsm_wait_ipacc_mdcx_ack_onenter,
+		.action = lchan_rtp_fsm_wait_ipacc_mdcx_ack,
+		.in_event_mask = 0
+			| S(LCHAN_RTP_EV_IPACC_MDCX_ACK)
+			| S(LCHAN_RTP_EV_IPACC_MDCX_NACK)
+			| S(LCHAN_RTP_EV_RELEASE)
+			| S(LCHAN_RTP_EV_ROLLBACK)
+			,
+		.out_state_mask = 0
+			| S(LCHAN_RTP_ST_READY)
+			| S(LCHAN_RTP_ST_ROLLBACK)
 			,
 	},
 	[LCHAN_RTP_ST_READY] = {
@@ -728,7 +750,7 @@ static const struct value_string lchan_rtp_fsm_event_names[] = {
 	{}
 };
 
-int lchan_rtp_fsm_timer_cb(struct osmo_fsm_inst *fi)
+static int lchan_rtp_fsm_timer_cb(struct osmo_fsm_inst *fi)
 {
 	struct gsm_lchan *lchan = lchan_rtp_fi_lchan(fi);
 	lchan->release.in_error = true;
@@ -737,7 +759,7 @@ int lchan_rtp_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	return 0;
 }
 
-void lchan_rtp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
+static void lchan_rtp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct gsm_lchan *lchan = lchan_rtp_fi_lchan(fi);
 	if (lchan->mgw_endpoint_ci_bts) {
