@@ -55,7 +55,6 @@ class TestCtrlBase(unittest.TestCase):
         except OSError:
             print("Current directory: %s" % os.getcwd(), file=sys.stderr)
             print("Consider setting -b", file=sys.stderr)
-        time.sleep(2)
 
         appstring = self.ctrl_app()[2]
         appport = self.ctrl_app()[0]
@@ -258,12 +257,12 @@ class TestCtrlBSC(TestCtrlBase):
         self.assertEqual(r['mtype'], 'SET_REPLY')
         self.assertEqual(r['var'], 'bts.0.trx.0.max-power-reduction')
         self.assertEqual(r['value'], '22')
-        
+
         r = self.do_get('bts.0.trx.0.max-power-reduction')
         self.assertEqual(r['mtype'], 'GET_REPLY')
         self.assertEqual(r['var'], 'bts.0.trx.0.max-power-reduction')
         self.assertEqual(r['value'], '22')
-        
+
         r = self.do_set('bts.0.trx.0.max-power-reduction', '1')
         self.assertEqual(r['mtype'], 'ERROR')
         self.assertEqual(r['error'], 'Value must be even')
@@ -292,7 +291,7 @@ class TestCtrlBSC(TestCtrlBase):
         r = self.do_get('bts.0.rf_state')
         self.assertEqual(r['mtype'], 'GET_REPLY')
         self.assertEqual(r['var'], 'bts.0.rf_state')
-        self.assertEqual(r['value'], 'inoperational,unlocked,on')
+        self.assertEqual(r['value'], 'inoperational,locked,on')
 
         r = self.do_set('rf_locked', '1')
         self.assertEqual(r['mtype'], 'SET_REPLY')
@@ -321,7 +320,7 @@ class TestCtrlBSC(TestCtrlBase):
         r = self.do_get('bts.0.rf_state')
         self.assertEqual(r['mtype'], 'GET_REPLY')
         self.assertEqual(r['var'], 'bts.0.rf_state')
-        self.assertEqual(r['value'], 'inoperational,unlocked,on')
+        self.assertEqual(r['value'], 'inoperational,locked,on')
 
         r = self.do_get('rf_locked')
         self.assertEqual(r['mtype'], 'GET_REPLY')
@@ -489,11 +488,47 @@ class TestCtrlBSC(TestCtrlBase):
         self.assertEqual(r['var'], 'mcc')
         self.assertEqual(r['value'], '002')
 
-def add_bsc_test(suite, workdir):
+class TestCtrlBSCNeighbor(TestCtrlBase):
+
+    def tearDown(self):
+        TestCtrlBase.tearDown(self)
+        os.unlink("tmp_dummy_sock")
+
+    def ctrl_command(self):
+        return ["./src/osmo-bsc/osmo-bsc", "-r", "tmp_dummy_sock", "-c",
+                "tests/ctrl/osmo-bsc-neigh-test.cfg"]
+
+    def ctrl_app(self):
+        return (4248, "./src/osmo-bsc/osmo-bsc", "OsmoBSC", "bsc")
+
+    def testCtrlNeighborResolutionLocalBtsNr(self):
+        r = self.do_get('neighbor_resolve_cgi_ps_from_lac_ci.1.123.871.63')
+        self.assertEqual(r['mtype'], 'GET_REPLY')
+        self.assertEqual(r['var'], 'neighbor_resolve_cgi_ps_from_lac_ci.1.123.871.63')
+        self.assertEqual(r['value'], '001-01-1-5-6969')
+
+    def testCtrlNeighborResolutionLocalWithoutArfcnBsic(self):
+        r = self.do_get('neighbor_resolve_cgi_ps_from_lac_ci.1.6969.880.55')
+        self.assertEqual(r['mtype'], 'GET_REPLY')
+        self.assertEqual(r['var'], 'neighbor_resolve_cgi_ps_from_lac_ci.1.6969.880.55')
+        self.assertEqual(r['value'], '001-01-1-6-123')
+
+    def testCtrlNeighborResolutionWrongSyntax(self):
+        r = self.do_get('neighbor_resolve_cgi_ps_from_lac_ci')
+        self.assertEqual(r['mtype'], 'ERROR')
+        self.assertEqual(r['error'], 'The format is <src_lac>,<src_cell_id>,<dst_arfcn>,<dst_bsic>')
+
+    def testCtrlNeighborResolutionRemote(self):
+        r = self.do_get('neighbor_resolve_cgi_ps_from_lac_ci.1.6969.23.32')
+        self.assertEqual(r['mtype'], 'GET_REPLY')
+        self.assertEqual(r['var'], 'neighbor_resolve_cgi_ps_from_lac_ci.1.6969.23.32')
+        self.assertEqual(r['value'], '023-42-423-2-5')
+
+def add_bsc_test(suite, workdir, klass):
     if not os.path.isfile(os.path.join(workdir, "src/osmo-bsc/osmo-bsc")):
         print("Skipping the BSC test")
         return
-    test = unittest.TestLoader().loadTestsFromTestCase(TestCtrlBSC)
+    test = unittest.TestLoader().loadTestsFromTestCase(klass)
     suite.addTest(test)
 
 if __name__ == '__main__':
@@ -526,6 +561,7 @@ if __name__ == '__main__':
     os.chdir(workdir)
     print("Running tests for specific control commands")
     suite = unittest.TestSuite()
-    add_bsc_test(suite, workdir)
+    add_bsc_test(suite, workdir, TestCtrlBSC)
+    add_bsc_test(suite, workdir, TestCtrlBSCNeighbor)
     res = unittest.TextTestRunner(verbosity=verbose_level).run(suite)
     sys.exit(len(res.errors) + len(res.failures))
