@@ -29,6 +29,7 @@
 #include <osmocom/bsc/arfcn_range_encode.h>
 #include <osmocom/bsc/system_information.h>
 #include <osmocom/bsc/abis_rsl.h>
+#include <osmocom/bsc/bts.h>
 
 #include <osmocom/core/application.h>
 #include <osmocom/core/byteswap.h>
@@ -122,7 +123,8 @@ static inline void _bts_uarfcn_add(struct gsm_bts *bts, uint16_t arfcn, uint16_t
 #define bts_init(net) _bts_init(net, __func__)
 static inline struct gsm_bts *_bts_init(struct gsm_network *net, const char *msg)
 {
-	struct gsm_bts *bts = gsm_bts_alloc(net, 0);
+	struct gsm_bts_sm *bts_sm = gsm_bts_sm_alloc(net, 0);
+	struct gsm_bts *bts = bts_sm->bts[0];
 	if (!bts) {
 		printf("BTS allocation failure in %s()\n", msg);
 		exit(1);
@@ -139,8 +141,10 @@ static inline void _bts_del(struct gsm_bts *bts, const char *msg)
 {
 	osmo_stat_item_group_free(bts->bts_statg);
 	rate_ctr_group_free(bts->bts_ctrs);
+	if (osmo_timer_pending(&bts->acc_mgr.rotate_timer))
+		osmo_timer_del(&bts->acc_mgr.rotate_timer);
 	/* no need to llist_del(&bts->list), we never registered the bts there. */
-	talloc_free(bts);
+	talloc_free(bts->site_mgr);
 	printf("BTS deallocated OK in %s()\n", msg);
 }
 
@@ -269,34 +273,6 @@ static inline void test_si2q_long(struct gsm_network *net)
 	bts_del(bts);
 }
 
-static void test_mi_functionality(void)
-{
-	const char *imsi_odd  = "987654321098763";
-	const char *imsi_even = "9876543210987654";
-	const uint32_t tmsi = 0xfabeacd0;
-	uint8_t mi[128];
-	unsigned int mi_len;
-	char mi_parsed[GSM48_MI_SIZE];
-
-	printf("Testing parsing and generating TMSI/IMSI\n");
-
-	/* tmsi code */
-	mi_len = gsm48_generate_mid_from_tmsi(mi, tmsi);
-	gsm48_mi_to_string(mi_parsed, sizeof(mi_parsed), mi + 2, mi_len - 2);
-	COMPARE((uint32_t)strtoul(mi_parsed, NULL, 10), ==, tmsi);
-
-	/* imsi code */
-	mi_len = gsm48_generate_mid_from_imsi(mi, imsi_odd);
-	gsm48_mi_to_string(mi_parsed, sizeof(mi_parsed), mi + 2, mi_len -2);
-	printf("hex: %s\n", osmo_hexdump(mi, mi_len));
-	COMPARE_STR(mi_parsed, imsi_odd);
-
-	mi_len = gsm48_generate_mid_from_imsi(mi, imsi_even);
-	gsm48_mi_to_string(mi_parsed, sizeof(mi_parsed), mi + 2, mi_len -2);
-	printf("hex: %s\n", osmo_hexdump(mi, mi_len));
-	COMPARE_STR(mi_parsed, imsi_even);
-}
-
 struct {
 	int range;
 	int arfcns_num;
@@ -924,8 +900,6 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
-	test_mi_functionality();
-
 	test_si_range_helpers();
 	test_arfcn_filter();
 	test_print_encoding();
@@ -979,8 +953,8 @@ void gscon_submit_rsl_dtap(struct gsm_subscriber_connection *conn,
 
 bool lchan_may_receive_data(struct gsm_lchan *lchan) { return  true; }
 
-int bsc_compl_l3(struct gsm_subscriber_connection *conn, struct msgb *msg,
-		 uint16_t chosen_channel) { return 0; }
+int bsc_compl_l3(struct gsm_lchan *lchan, struct msgb *msg, uint16_t chosen_channel)
+{ return 0; }
 
 void bsc_dtap(struct gsm_subscriber_connection *conn, uint8_t link_id,
 	      struct msgb *msg) {}
@@ -998,3 +972,8 @@ int rsl_data_request(struct msgb *msg, uint8_t link_id) { return 0; }
 int rsl_encryption_cmd(struct msgb *msg) { return 0; }
 
 int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan) { return 0; }
+
+void pcu_info_update(struct gsm_bts *bts) {};
+int rsl_sacch_filling(struct gsm_bts_trx *trx, uint8_t type, const uint8_t *data, int len) { return 0; }
+int rsl_bcch_info(const struct gsm_bts_trx *trx, enum osmo_sysinfo_type si_type, const uint8_t *data, int len)
+{ return 0; }
